#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <time.h>
#include <string.h>

#include "CDVDiso.h"
#include "Config.h"

char *methods[] = {
	".Z  - compress faster",
//	".bz - compress better",
	NULL
};

char *LibName = "Linuzappz Iso CDVD Driver";

const unsigned char version = PS2E_CDVD_VERSION;
const unsigned char revision = 0;
const unsigned char build = 5;

char* CALLBACK PS2EgetLibName() {
	return LibName;
}

u32 CALLBACK PS2EgetLibType() {
	return PS2E_LT_CDVD;
}

u32 CALLBACK PS2EgetLibVersion2(u32 type) {
	return (version << 16) | (revision << 8) | build;
}

#ifdef CDVD_LOG
void __Log(char *fmt, ...) {
	va_list list;

	va_start(list, fmt);
	vfprintf(cdvdLog, fmt, list);
	va_end(list);
}
#endif


s32 CALLBACK CDVDinit() {
#ifdef CDVD_LOG
	cdvdLog = fopen("logs/cdvdLog.txt", "w");
	if (cdvdLog == NULL) {
		cdvdLog = fopen("cdvdLog.txt", "w");
		if (cdvdLog == NULL) {
			SysMessage("Can't create cdvdLog.txt"); return -1;
		}
	}
	setvbuf(cdvdLog, NULL,  _IONBF, 0);
	CDVD_LOG("CDVDinit\n");
#endif
	memset(cdIso, 0, sizeof(cdIso));
	return 0;
}

void CALLBACK CDVDshutdown() {
#ifdef CDVD_LOG
	fclose(cdvdLog);
#endif
}

s32 CALLBACK CDVDopen() {
	LoadConf();

	if (*IsoFile == 0) {
		char temp[256];

		CfgOpenFile();

		LoadConf();
		strcpy(temp, IsoFile);
		*IsoFile = 0;
		SaveConf();
		strcpy(IsoFile, temp);
	}

	iso = isoOpen(IsoFile);
	if (iso == NULL) {
		SysMessage("Error loading %s\n", IsoFile);
		return -1;
	}

	if (iso->type == ISOTYPE_DVD) {
		 cdtype = CDVD_TYPE_PS2DVD;
	} else 
	if (iso->type == ISOTYPE_AUDIO) {
		 cdtype = CDVD_TYPE_CDDA;
	} else {
		cdtype = CDVD_TYPE_PS2CD;
	}

	if (BlockDump) {
		fdump = isoCreate("block.dump", ISOFLAGS_BLOCKDUMP);
		if (fdump) {
			isoSetFormat(fdump, iso->blockofs, iso->blocksize, iso->blocks);
		}
	} else {
		fdump = NULL;
	}

	return 0;
}

void CALLBACK CDVDclose() {
	isoClose(iso);
	if (fdump != NULL) {
		isoClose(fdump);
	}
}

s32 CALLBACK CDVDgetTN(cdvdTN *Buffer) {
	Buffer->strack = 1;
	Buffer->etrack = 1;

	return 0;
}

s32 CALLBACK CDVDgetTD(u8 Track, cdvdTD *Buffer) {
	if (Track == 0) {
		Buffer->lsn = iso->blocks;
	} else {
		Buffer->type = CDVD_MODE1_TRACK;
		Buffer->lsn = 0;
	}

	return 0;
}

s32 CALLBACK CDVDreadTrack(u32 lsn, int mode) {
	int _lsn = lsn;
	if (_lsn < 0) {
//		lsn = 2097152 + (-_lsn);
		lsn = iso->blocks - (-_lsn);
	}
//	printf ("CDRreadTrack %d\n", lsn);

	isoReadBlock(iso, cdbuffer, lsn);
	if (fdump != NULL) {
		isoWriteBlock(fdump, cdbuffer, lsn);
	}

	pbuffer = cdbuffer;
	switch (mode) {
		case CDVD_MODE_2352: break;
		case CDVD_MODE_2340: pbuffer+= 12; break;
		case CDVD_MODE_2328: pbuffer+= 24; break;
		case CDVD_MODE_2048: pbuffer+= 24; break;
	}

	return 0;
}

u8* CALLBACK CDVDgetBuffer() {
	return pbuffer;
}

s32  CALLBACK CDVDgetType() {
	return cdtype;
}

s32  CALLBACK CDVDgetTrayStatus() {
	return CDVD_TRAY_CLOSE;
}

s32 CALLBACK CDVDtest() {
	if (*IsoFile == 0)
		return 0;

	iso = isoOpen(IsoFile);
	if (iso == NULL) return -1;
	isoClose(iso);

	return 0;
}

