/*  FireWire 
 *  Copyright (C) 2004 PCSX2 Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>
#include <errno.h>

#include "FW.h"

const unsigned char version  = PS2E_FIREWIRE_VERSION;
const unsigned char revision = 0;
const unsigned char build    = 1;    // increase that with each version

static char *libraryName     = "FWlinuz Driver";

s8 *fwregs;
u8 phyregs[16];

#define fwRs32(mem)	(*(s32*)&fwregs[(mem) & 0xffff])
#define fwRu32(mem)	(*(u32*)&fwregs[(mem) & 0xffff])


u32 CALLBACK PS2EgetLibType() {
	return PS2E_LT_FIREWIRE;
}

char* CALLBACK PS2EgetLibName() {
	return libraryName;
}

u32 CALLBACK PS2EgetLibVersion2(u32 type) {
	return (version<<16) | (revision<<8) | build;
}

void __Log(char *fmt, ...) {
	va_list list;

	if (!conf.Log) return;

	va_start(list, fmt);
	vfprintf(fwLog, fmt, list);
	va_end(list);
}

s32 CALLBACK FireWireinit() {
    LoadConfig();
#ifdef FW_LOG
	fwLog = fopen("logs/fwLog.txt", "w");
	setvbuf(fwLog, NULL,  _IONBF, 0);
	FW_LOG("FWlinuz plugin version %d,%d\n",revision,build);
	FW_LOG("FW init\n");
#endif

	fwregs = (s8*)malloc(0x10000);
	if (fwregs == NULL) {
		SysMessage("Error allocating Memory\n"); return -1;
	}
	memset(phyregs, 0, sizeof(phyregs));
	phyregs[0] = 0xc1;

	return 0;
}

void CALLBACK FireWireshutdown() {
	free(fwregs);

#ifdef FW_LOG
	fclose(fwLog);
#endif
}

s32 CALLBACK FireWireopen(void *pDsp) {
#ifdef FW_LOG
	FW_LOG("FireWire open\n");
#endif

	return 0;
}

void CALLBACK FireWireclose() {
}




u32  CALLBACK FireWireread32(u32 addr) {
	u32 ret=0;

	switch (addr) {
		case 0x1f808400:
			ret = (fwRu32(addr) & 0xffc00000) | 0x1;
			break;

		case 0x1f808410:
			ret = 0x8;
			break;

		default:
			ret = fwRu32(addr);
			break;
	}
#ifdef FW_LOG
	FW_LOG("FW read mem 0x%x: 0x%x\n", addr, ret);
#endif	

	return ret;
}

void _irq0(int num) {
	fwRu32(0x8420)|= 1<<num;
	if (fwRu32(0x8420) & fwRu32(0x8424)) {
		FWirq();
	}
}

void phy_read() {
	int reg = (PHYACC >> 24) & 0xf;

	printf("phy read  0x%x: 0x%2.2x\n", reg, phyregs[reg]);
	PHYACC = (PHYACC & ~0xff) | phyregs[reg];
	PHYACC = (PHYACC & ~0xf00) | (reg << 8);
	PHYACC&= ~0x80000000;
	_irq0(30);
}

void phy_write() {
	int reg = (PHYACC >> 24) & 0xf;
	u8 value = PHYACC & 0xff;

	printf("phy write 0x%x: 0x%2.2x\n", reg, value);
	switch (reg) {
		case 0x0: break;
		case 0x2: break;
		case 0x3: break;
		case 0x4: phyregs[reg] = (phyregs[reg] & 0x1c) | (value & ~0x1c); break;
		case 0x6: break;
		default: phyregs[reg] = value;
	}

	PHYACC&= ~0x40000000;
	_irq0(31);
}

void CALLBACK FireWirewrite32(u32 addr, u32 value) {
	switch (addr) {
		case 0x1f808414:
			fwRu32(addr) = value;
			if (value & 0x80000000) phy_read();
			if (value & 0x40000000) phy_write();
			break;

		case 0x1f808420:
		case 0x1f808428:
		case 0x1f808430:
			fwRu32(addr) = (fwRu32(addr) & ~value);
			break;

		default:
			fwRu32(addr) = value;
			break;
	}
#ifdef FW_LOG
	FW_LOG("FW write mem 0x%x: 0x%x\n", addr, value);
#endif
}

void CALLBACK FireWireirqCallback(void (*callback)()) {
	FWirq = callback;
}

s32 CALLBACK FireWirefreeze(int mode, freezeData *data) {
	return 0;
}

s32  CALLBACK FireWiretest() {
	return 0;
}

